package com.mgs.demo.util;

import cn.edu.sjtu.loccs.bouncycastle.asn1.ASN1InputStream;
import cn.edu.sjtu.loccs.bouncycastle.asn1.ASN1Sequence;
import cn.edu.sjtu.loccs.bouncycastle.asn1.DERBitString;
import cn.edu.sjtu.loccs.bouncycastle.crypto.digests.SM3Digest;
import cn.edu.sjtu.loccs.bouncycastle.jcajce.provider.asymmetric.sm2.SM2ParameterSpec;
import cn.edu.sjtu.loccs.bouncycastle.jcajce.provider.asymmetric.sm2.SM2PublicKey;
import cn.edu.sjtu.loccs.bouncycastle.jcajce.provider.asymmetric.sm2.SM2PublicKeySpec;
import cn.edu.sjtu.loccs.bouncycastle.jcajce.provider.asymmetric.sm2.curve.SM2NamedCurveTable;
import cn.edu.sjtu.loccs.bouncycastle.openssl.PEMParser;
import cn.edu.sjtu.loccs.bouncycastle.util.io.pem.PemObject;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.lang3.StringUtils;
import org.bouncycastle.jce.provider.BouncyCastleProvider;

import javax.servlet.http.HttpServletRequest;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.security.KeyFactory;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.Security;
import java.security.Signature;
import java.security.spec.X509EncodedKeySpec;
import java.util.Base64;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by scott on 2018/12/27.
 */
public class HttpSignUtil {

    public static final String LF = "\n";
    public static final String MD5_KEY = "MGS-SIGN-SECRET-KEY-MD5";
    public static final String RSA_KEY = "MGS-SIGN-SECRET-KEY-RSA";
    public static final String SM2_KEY = "MGS-SIGN-SECRET-KEY-SM2";
    public static final String SM3_KEY = "MGS-SIGN-SECRET-KEY-SM3";

    private static final String md5Key = "eae459a730f9ba26d12f3dd3a8c95220";
    private static final String rsaPubKey = "MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCz09z6e9WOcNq+nUMX8Vq1Xe2EmJxuR3Xbture6z0z0n+Mr+bssDOyyMQqvEzgSNaTQSD2UlWaaO6OY8yu8+3VL8YjD+3Slrx94JIMCz5YHuWEFc8HY4T5R6MIfumNKql9v0natqehyx5dawTpg3LvKTl2naBb3l9lxZ7t9pm+NQIDAQAB";
    private static final String sm3Key = "eae459a730f9ba26d12f3dd3a8c95220";
    private static final String sm2PubKey = "-----BEGIN PUBLIC KEY-----\n" +
            "MFkwEwYHKoZIzj0CAQYIKoEcz1UBgi0DQgAEEg+mPzlW9iSMAOMUkhe+UZnDm8zX\n" +
            "+56tDrzdoOneNz9k1OvfaPxijM9sui/nYCpRYyd2F/gSUTGkb4iIabGIvw==\n" +
            "-----END PUBLIC KEY-----";


    private static final Map<String, Object> KEY_MAP = new ConcurrentHashMap<>();

    static {
        if (Security.getProvider("LOCCSBC") == null) {
            //LoccsBCUtil.addLoccsBCProvider();
            Security.addProvider(new BouncyCastleProvider());
        }

        initKeyMap();
    }


    private static void initKeyMap() {
        KEY_MAP.put(MD5_KEY, md5Key);
        KEY_MAP.put(RSA_KEY, buildRsaPubKey(rsaPubKey));
        KEY_MAP.put(SM3_KEY, sm3Key);
        KEY_MAP.put(SM2_KEY, buildSm2PubKey(sm2PubKey));
    }

    private static PublicKey buildRsaPubKey(String rsaPubKey) {
        try {
            PublicKey pubKey = getPublicKeyFromX509("RSA", rsaPubKey.getBytes());
            return pubKey;
        } catch (Throwable t) {
            t.printStackTrace();
        }
        return null;
    }

    private static SM2PublicKey buildSm2PubKey(String sm2PubKey) {
        try {
            PEMParser reader = new PEMParser(new StringReader(sm2PubKey));
            PemObject pemObject = reader.readPemObject();
            reader.close();

            ASN1InputStream asn1In = new ASN1InputStream(pemObject.getContent());
            ASN1Sequence derSequence = (ASN1Sequence) asn1In.readObject();
            asn1In.close();

            Object obj = derSequence.getObjectAt(1);
            if (!(obj instanceof DERBitString)) {
                throw new Exception("pubkey parse error");
            }
            DERBitString derBitString = (DERBitString) obj;
            SM2ParameterSpec spec = SM2NamedCurveTable.getParameterSpec("sm2p256v1", "1234567812345678");
            SM2PublicKeySpec pubSpec = new SM2PublicKeySpec(spec.getCurve().decodePoint(derBitString.getBytes()), spec);
            KeyFactory keyFactory = KeyFactory.getInstance("SM2", "LOCCSBC");
            SM2PublicKey publicKey = (SM2PublicKey) keyFactory.generatePublic(pubSpec);
            return publicKey;
        } catch (Throwable t) {
            t.printStackTrace();
        }

        return null;
    }

    public static boolean checkSign(HttpServletRequest req) throws Throwable {
        String key = req.getHeader("X-Mgs-Proxy-Signature-Secret-Key");
        switch (key) {
            case MD5_KEY:
                return checkSignMd5(req);
            case RSA_KEY:
                return checkSignRsa(req);
            case SM2_KEY:
                return checkSignSm2(req);
            case SM3_KEY:
                return checkSignSm3(req);
            default:
                return false;
        }
    }

    /**
     * 采用 SM2 算法验签
     */
    public static boolean checkSignSm2(HttpServletRequest req) throws Throwable {
        String sign = req.getHeader("X-Mgs-Proxy-Signature");//移动网关传过来的签名
        SM2PublicKey publicKey = (SM2PublicKey) KEY_MAP.get(req.getHeader("X-Mgs-Proxy-Signature-Secret-Key"));

        String stringToSign = buildStringToSign(req);

        Signature signature = Signature.getInstance("SM3withSM2", "LOCCSBC");
        signature.initVerify(publicKey);
        signature.update(stringToSign.getBytes("UTF-8"));
        return signature.verify(Hex.decodeHex(sign));

    }


    /**
     * 采用 SM3 算法验签
     */
    public static boolean checkSignSm3(HttpServletRequest req) throws Throwable {
        boolean isSignLegal = false;
        String sign = req.getHeader("X-Mgs-Proxy-Signature");//移动网关传过来的签名
        String salt = (String) KEY_MAP.get(req.getHeader("X-Mgs-Proxy-Signature-Secret-Key"));     //sm3 salt
        String stringToSign = buildStringToSign(req);
        String toSignedContent = stringToSign + salt;

        byte[] srcData = toSignedContent.getBytes("UTF-8");
        SM3Digest digest = new SM3Digest();
        digest.update(srcData, 0, srcData.length);
        byte[] resultHash = new byte[digest.getDigestSize()];
        digest.doFinal(resultHash, 0);
        String computedSign = Hex.encodeHexString(resultHash);
        isSignLegal = sign.equals(computedSign);

        return isSignLegal;

    }

    /**
     * 采用 MD5 算法验签
     */
    public static boolean checkSignMd5(HttpServletRequest req) throws Throwable {
        boolean isSignLegal = false;
        String sign = req.getHeader("X-Mgs-Proxy-Signature");//移动网关传过来的签名
        String salt = (String) KEY_MAP.get(req.getHeader("X-Mgs-Proxy-Signature-Secret-Key"));     //MD5 Salt

        String stringToSign = buildStringToSign(req);

        MessageDigest digest = MessageDigest.getInstance("MD5");
        String toSignedContent = stringToSign + salt;
        byte[] content = digest.digest(toSignedContent.getBytes("utf8"));
        String computedSign = new String(Hex.encodeHexString(content));
        isSignLegal = sign.equals(computedSign) ? true : false;

        return isSignLegal;
    }

    /**
     * 采用 RSA 算法验签
     */
    public static boolean checkSignRsa(HttpServletRequest req) throws Throwable {
        boolean isSignLegal = false;
        String sign = req.getHeader("X-Mgs-Proxy-Signature");//移动网关传过来的签名
        PublicKey publicKey = (PublicKey) KEY_MAP.get(req.getHeader("X-Mgs-Proxy-Signature-Secret-Key"));  //移动网关的 RSA 公钥

        String stringToSign = buildStringToSign(req);

        Signature signature = Signature.getInstance("SHA1WithRSA");
        signature.initVerify(publicKey);
        signature.update(stringToSign.getBytes("utf8"));
        isSignLegal = signature.verify(Base64.getDecoder().decode(sign.getBytes("utf8")));

        return isSignLegal;
    }

    /**
     * 组织加签数据
     */
    private static String buildStringToSign(HttpServletRequest req) throws IOException, NoSuchAlgorithmException {
        String httpMethod = req.getMethod().toUpperCase();
        String url = buildUrl(req);
        String contentMD5 = buildContentMd5(req);

        StringBuilder stringToSign = new StringBuilder();
        stringToSign.append(httpMethod).append(LF);
        if (StringUtils.isNotBlank(contentMD5)) { //Content-MD5 需要判断是否为空，如果为空则跳过，但是为空也需要添加换行符 "\n"
            stringToSign.append(contentMD5);
        }
        stringToSign.append(LF);
        stringToSign.append(url);
        return stringToSign.toString();
    }

    private static String buildContentMd5(HttpServletRequest req) throws IOException, NoSuchAlgorithmException {
        String method = req.getMethod();
        String contentType = req.getContentType().toLowerCase();
        if ((StringUtils.equalsIgnoreCase(method, "POST") || StringUtils.equalsIgnoreCase(method, "PUT"))
                && !contentType.contains("application/x-www-form-urlencoded")) {
            BufferedReader br = req.getReader();
            String str = "";
            String reqBody = "";
            while ((str = br.readLine()) != null) {
                reqBody += str;
            }
            if (StringUtils.isBlank(reqBody)) {
                reqBody = "null";
            }
            byte[] contentBytes = reqBody.getBytes("utf8");
            final MessageDigest md = MessageDigest.getInstance("MD5");
            md.reset();
            md.update(contentBytes);
            return new String(Base64.getEncoder().encode(md.digest())).trim();
        } else {
            return null;
        }
    }

    private static String buildUrl(HttpServletRequest req) {
        String path = req.getServletPath();

        Map<String, String[]> parameterMap = req.getParameterMap();
        TreeMap<String, Object> params = new TreeMap<>();
        for (String key : parameterMap.keySet()) {
            params.put(key, parameterMap.get(key)[0]);
        }
        StringBuilder builder = new StringBuilder();
        // uri
        builder.append(path);
        // 有Query+Form参数
        if (params.size() > 0) {
            builder.append('?');
            builder.append(buildUrlParam(params));
        }
        return builder.toString();
    }

    private static String buildUrlParam(Map<String, Object> paramMap) {
        StringBuilder builder = new StringBuilder();
        for (Map.Entry<String, Object> e : paramMap.entrySet()) {
            if (builder.length() > 0) {
                builder.append('&');
            }
            String key = e.getKey();
            Object value = e.getValue();
            if (value != null) {
                if (value instanceof List) {
                    List list = (List) value;
                    if (list.size() == 0) {
                        builder.append(key);
                    } else {
                        builder.append(key).append("=").append(String.valueOf(list.get(0)));
                    }
                } else if (value instanceof Object[]) {
                    Object[] objs = (Object[]) value;
                    if (objs.length == 0) {
                        builder.append(key);
                    } else {
                        builder.append(key).append("=").append(String.valueOf(objs[0]));
                    }
                } else {
                    builder.append(key).append("=").append(String.valueOf(value));
                }
            }
        }
        return builder.toString();
    }

    public static PublicKey getPublicKeyFromX509(String algorithm, byte[] encodedKeyBytes) throws Throwable {
        KeyFactory keyFactory = KeyFactory.getInstance(algorithm);
        byte[] encodedKey = Base64.getDecoder().decode(encodedKeyBytes);
        return keyFactory.generatePublic(new X509EncodedKeySpec(encodedKey));
    }
}
